@@ -72,6 +72,12 @@ class HashTable h where
     -- Returns the second part of the tuple returned by /f/.
     mutate :: (Eq k, Hashable k) =>
               h s k v -> k -> (Maybe v -> (Maybe v, a)) -> ST s a
+    mutate tbl k f = mutateST tbl k (pure . f)
+
+    -- | As 'mutate', except that the action can perform additional side
+    -- effects.
+    mutateST :: (Eq k, Hashable k) =>
+                h s k v -> k -> (Maybe v -> ST s (Maybe v, a)) -> ST s a
 
     -- | Inserts a key/value mapping into a hash table, replacing any existing
     -- mapping for that key.
